/* 
 * Utilities for synchronization with a relay
 * 
 * class VersionVector
 * functions has, isRelay, trimRelay
 */


export class VersionVector {
    
    // Map(<String peer, Number clk>)
    
    
    /**
     * @param {type} str A string of the form {peer1:v1,peer2:v2,peer3:v3} (JSON without double quotes)
     * @returns {undefined}
     */
    constructor(str = "{}") {
        // add double quotes around peer names
        const jsonStr = str.replace(/([^{,]+):/g,'"$1":');
        this.map = new Map(Object.entries(JSON.parse(jsonStr))); 
    }
    
    
    /**
     * Produces a string version of this version vector, of the form
     * {peer1:v1,peer2:v2,peer3:v3} (JSON without double quotes)
     * @returns {String}
     */
    toString() { 
        const JsonStr = JSON.stringify(Object.fromEntries(this.map));
        return JsonStr.replaceAll('"','');
    }
    
    
    
    equals(other) {
        if (this.getSize() !== other.getSize()) return false;
        for (const peer of this.getKeys()) {
            if (this.get(peer) !== other.get(peer)) return false;
        }
        return true;
    }
    
    
    getKeys() {
        return this.map.keys();
    }
    
    
    get(peer) {
        let vnum = this.map.get(peer);
        if (vnum === undefined) vnum = 0;
        return vnum;
    }
    
    
    set(peer, vnum) {
        this.map.set(peer, vnum);
    }
    
    
    /**
     * Gives the size of this vector
     * @returns {VersionVector.map.size}
     */
    getSize() {
        return this.map.size;
    }
        
        
    isEmpty() {
        return (this.getSize() === 0);
    }    
    
        
    /**
     * Produces a deep copy of this vector
     * @returns {VersionVector}
     */
    clone() {
        let result =  new VersionVector();
        for (const [peer, vnum] of this.map) {
            result.set(peer, vnum);
        }
        return result;
    }
    
    
    /**
     * Indicates if there exists a version number in this vector that is strictly lower than 
     * the corresponding version number (ie for the same peer) in another vector and 
     * no version number is strictly greater than the corresponding version number in this other vector.
     * @param {type} other
     * @returns {boolean}
     */
    before(other) {
        let bef = false;
        for (const peer of this.getKeys()) {
            const myVnum = this.get(peer);
            const otherVnum = other.get(peer);
            if (myVnum > otherVnum) return false;
            if (myVnum < otherVnum) bef = true;
        }
        if (bef) {
            return true;
        }
        else { // all version numbers are equal
            return (other.getSize() > this.getSize());
        } 
    }
    




    after(other) {
        return other.before(this);
    }
    
    
    
    
    /**
     * Overwrite this vector with the min vector between this vector and another one
     * @param {VersionVector} other
     * @returns {undefined}
     */
    min(other) {
        for (const peer of other.getKeys()) {
            const otherVnum = other.get(peer);
            const localVnum = this.get(peer);
            
            if (otherVnum < localVnum) {
                this.set(peer, otherVnum);
            } 
        }
    }
    
    
     /**
     * Overwrite this vector with the max vector between this vector and another one
     * @param {VersionVector} other
     * @returns {undefined}
     */
    max(other) {
        for (const peer of other.getKeys()) {
            const otherVnum = other.get(peer);
            const localVnum = this.get(peer);
            if (otherVnum > localVnum) {
                this.set(peer, otherVnum);
            } 
        }
    }
    
    
    /**
     * Indicates if there exists one peer in this vector for which the version number 
     * is strictly greater than the corresponding version number in the other vector
     * @param {VersionVector} other
     * @returns {boolean}
     */ 
    isOver(other) {
        for (const peer of this.getKeys()) {
            const localVnum = this.get(peer);
            const otherVnum = other.get(peer);
            if (localVnum > otherVnum) return true;
        }
        return false;
     }
    
    
    /**
     * Increments the version number of a peer in this vector
     * @param {String} peer The peer name
     * @returns {String}
     */
    inc(peer) {
        let clk = this.get(peer);
        clk++;
        this.set(peer, clk);
    }
    
    
    /**
     * Merge this vector with another one
     * @param {VersionVector} other
     * @returns {undefined}
     */
    merge(other) {
    	this.max(other);
    }
}





/// Adapted from below to Uint8array
// cyrb53 (c) 2018 bryc (github.com/bryc). License: Public domain. Attribution appreciated.
// A fast and simple 64-bit (or 53-bit) string hash function with decent collision resistance.
// Largely inspired by MurmurHash2/3, but with a focus on speed/simplicity.
// See https://stackoverflow.com/questions/7616461/generate-a-hash-from-string-in-javascript/52171480#52171480
// https://github.com/bryc/code/blob/master/jshash/experimental/cyrb53.js
const cyrb64 = (tab, seed = 0) => {
  let h1 = 0xdeadbeef ^ seed, h2 = 0x41c6ce57 ^ seed;
  for(let i = 0, ch; i < tab.length; i++) {
    ch = tab[i];
    h1 = Math.imul(h1 ^ ch, 2654435761);
    h2 = Math.imul(h2 ^ ch, 1597334677);
  }
  h1  = Math.imul(h1 ^ (h1 >>> 16), 2246822507);
  h1 ^= Math.imul(h2 ^ (h2 >>> 13), 3266489909);
  h2  = Math.imul(h2 ^ (h2 >>> 16), 2246822507);
  h2 ^= Math.imul(h1 ^ (h1 >>> 13), 3266489909);
  // For a single 53-bit numeric return value we could return
  // 4294967296 * (2097151 & h2) + (h1 >>> 0);
  // but we instead return the full 64-bit value:
  return [h2>>>0, h1>>>0];
};


/**
 * Compute a hash code for a an array of bytes
 * An improved, *insecure* 64-bit hash that's short, fast, and has no dependencies.
 * Output is always 14 digit (characters).
 * 
 *@param {Uint8array} tab
 *@param {Integer} seed
 *@returns {String} the hash code
 */
export function hash(tab, seed = 0) {
  const [h2, h1] = cyrb64(tab, seed);
  return h2.toString(36).padStart(7, '0') + h1.toString(36).padStart(7, '0');
};




/**
 * Indicates if a peer name is a relay
 * @param {String} peer
 * @returns {boolean}
 */
export function isRelay(peer) {
    return peer.startsWith("R_");
}


/**
 * Extracts the plain name from a relay peer name by removing the prefix "R_" if any.
 * @param {String} peer
 * @returns {String}
 */
export function trimRelay(peer) {
    if (isRelay(peer)) {
        return peer.substring(2);
    }
    else {
        return peer;
    }
}
